package org.expeditee.auth.mail;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.security.InvalidKeyException;
import java.security.Key;
import java.security.KeyFactory;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.cert.CertificateException;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.PKCS8EncodedKeySpec;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Base64;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Scanner;
import java.util.function.Supplier;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.SecretKey;

import org.expeditee.auth.AuthenticatorBrowser;
import org.expeditee.encryption.CryptographyConstants;
import org.expeditee.gui.FrameIO;
import org.expeditee.gui.MessageBay;
import org.expeditee.items.Text;
import org.expeditee.settings.identity.secrets.KeyList;

public class Mail implements CryptographyConstants {
	private String user;
	public static final SimpleDateFormat FORMAT = new SimpleDateFormat("EEE, d MMM yyyy HH:mm:ss Z");

	public Mail(String user) {
		this.user = user.toLowerCase();
	}
	
	public void sendMail(MailEntry mail, String colleagueName) {
		// Get database file to place mail into
		Path databaseFilePath = getDatabaseFilePath(colleagueName);

		// Obtain colleagues public key
		PublicKey publicKey = null;
		try {
			publicKey = AuthenticatorBrowser.getInstance().getPublicKey(colleagueName);
		} catch (InvalidKeySpecException | NoSuchAlgorithmException | KeyStoreException | CertificateException
				| ClassNotFoundException | IOException | SQLException e) {
			System.err.println("Error while sending message.  Unable to obtain public key for colleague "
					+ colleagueName + ".  Exception message: " + e.getMessage());
			return;
		}

		// Check we got public key
		if (publicKey == null) {
			System.err.println("Error while sending message.  " + "Unable to obtain public key for colleague.  "
					+ "Have you exchanged contact details?");
			return;
		}

		// Send message
		MailEntry encryptedMailEntryForSending = mail.encryptMailEntry(publicKey, () -> {
			try {
				return Cipher.getInstance(AsymmetricAlgorithm + AsymmetricAlgorithmParameters);
			} catch (NoSuchAlgorithmException | NoSuchPaddingException e) {
				e.printStackTrace();
				return null;
			}
		}, "");
		encryptedMailEntryForSending.writeToMailDatabase(databaseFilePath);

		// Do it all again if demo mode is on (ensures sync without having to move db
		// files)
		if (Boolean.getBoolean("expeditee.demo-mode")) {
			databaseFilePath = getDatabaseFilePath(colleagueName, true);
			encryptedMailEntryForSending.writeToMailDatabase(databaseFilePath);
		}
	}

	public void sendOneOffMail(MailEntry mail, String colleagueName, SecretKey key) {
		// Get database file to place mail into
		Path databaseFilePath = getDatabaseFilePath(colleagueName);

		// Send message
		MailEntry encryptedMailEntryForSending = mail.encryptMailEntry(key, () -> {
			try {
				return Cipher.getInstance(SymmetricAlgorithm + SymmetricAlgorithmParameters);
			} catch (NoSuchAlgorithmException | NoSuchPaddingException e) {
				e.printStackTrace();
				return null;
			}
		}, MailEntry.SINGLE_USE_SENDER_TAG);
		encryptedMailEntryForSending.writeToMailDatabase(databaseFilePath);

		// Do it all again if demo mode is on (ensures sync without having to move db
		// files)
		if (Boolean.getBoolean("expeditee.demo-mode")) {
			databaseFilePath = getDatabaseFilePath(colleagueName, true);
			encryptedMailEntryForSending.writeToMailDatabase(databaseFilePath);
		}
	}

	public List<MailEntry> checkMail() {
		PrivateKey key = getPrivateKey();
		List<Path> databasePaths = getDatabaseInboxFiles();
		List<MailEntry> receivedMailEncrypted = loadMailFromDatabases(databasePaths);
		List<MailEntry> receivedMail = decryptMailEntries(receivedMailEncrypted, key);
		// TODO: delete processed messages from database files
		touchLastAccessedFiles(databasePaths);
		return receivedMail;
	}

	public class MailEntry {
		private static final String SINGLE_USE_SENDER_TAG = "=";
		private String timestamp;
		private String sender;
		private String receiver;
		private String subject;
		private String message;
		private Map<String, String> options;
		private Path source;

		public MailEntry(String timestamp, String sender, String receiver, String subject, String message,
				Map<String, String> options) {
			this.timestamp = timestamp;
			this.sender = sender;
			this.receiver = receiver;
			this.subject = subject;
			this.message = message;
			this.options = options;
		}

		//@formatter:off
		public String getTimestamp() { return timestamp; }
		public String getSender() { return sender; }
		public String getReceiver() { return receiver; }
		public String getSubject() { return subject; }
		public String getMessage() { return message; }
		public Map<String, String> getOptionsTextActionMap() { return options; }
		public Path getDatabaseSource() { return source; }
		public void setDatabaseSource(Path source) { this.source = source; }
		public boolean isSingleUseEncryption() {
			return sender.startsWith(MailEntry.SINGLE_USE_SENDER_TAG);
		}
		//@formatter:on

		private MailEntry encryptMailEntry(Key key, Supplier<Cipher> generateCipher, String markSender) {
			if (markSender == null)
				markSender = "";
			try {
				// Encrypt sender, receiver, subject and message
				Cipher cipher = generateCipher.get();
				cipher.init(Cipher.ENCRYPT_MODE, key);
				String sender = markSender + Base64.getEncoder().encodeToString(cipher.doFinal(getSender().toLowerCase().getBytes()));
				cipher.init(Cipher.ENCRYPT_MODE, key);
				String receiver = Base64.getEncoder().encodeToString(cipher.doFinal(getReceiver().getBytes()));
				cipher.init(Cipher.ENCRYPT_MODE, key);
				String subject = Base64.getEncoder().encodeToString(cipher.doFinal(getSubject().getBytes()));
				cipher.init(Cipher.ENCRYPT_MODE, key);
				String message = Base64.getEncoder().encodeToString(cipher.doFinal(getMessage().getBytes()));

				// Encrypt reply options to be provided to receiver
				Map<String, String> optionsTextActionMap = new HashMap<String, String>();
				Map<String, String> toEncOptionsTextActionMap = getOptionsTextActionMap();
				for (String text : toEncOptionsTextActionMap.keySet()) {
					cipher.init(Cipher.ENCRYPT_MODE, key);
					String textEncrypted = Base64.getEncoder().encodeToString(cipher.doFinal(text.getBytes()));
					cipher.init(Cipher.ENCRYPT_MODE, key);
					String toEncAction = toEncOptionsTextActionMap.get(text);
					String actionEncrypted = Base64.getEncoder().encodeToString(cipher.doFinal(toEncAction.getBytes()));
					optionsTextActionMap.put(textEncrypted, actionEncrypted);
				}

				return new MailEntry(getTimestamp(), sender, receiver, subject, message, optionsTextActionMap);
			} catch (InvalidKeyException | IllegalBlockSizeException | BadPaddingException e) {
				e.printStackTrace();
				return null;
			}
		}

		private void writeToMailDatabase(Path databaseFile) {
			try {
				Connection c = DriverManager.getConnection("jdbc:sqlite:" + databaseFile);
				String sql = "INSERT INTO EXPMAIL (TIME,SND,REC,SUB,MSG,OPTS,OPTSVAL) VALUES (?, ?, ?, ?, ?, ?, ?);";
				PreparedStatement statement = c.prepareStatement(sql);
				statement.setString(1, getTimestamp());
				statement.setString(2, getSender());
				statement.setString(3, getReceiver());
				statement.setString(4, getSubject());
				statement.setString(5, message);
				String opts = Arrays.toString(getOptionsTextActionMap().keySet().toArray());
				statement.setString(6, opts);
				String optsVal = Arrays.toString(getOptionsTextActionMap().values().toArray());
				statement.setString(7, optsVal);
				statement.execute();
				statement.close();
			} catch (SQLException e) {
				e.printStackTrace();
			}
		}

		private MailEntry decryptMailEntry(Key key, Supplier<Cipher> generateCipher, String senderMarkFilter) {
			Cipher cipher = generateCipher.get();
			try {
				// Confirm this is a message is intended for this recipient
				cipher.init(Cipher.DECRYPT_MODE, key);
				byte[] receiverBytes = Base64.getDecoder().decode(getReceiver());
				String receiver = new String(cipher.doFinal(receiverBytes));
				if (!receiver.toLowerCase().equals(user)) {
					return null;
				}

				// Decrypt remainder of message
				cipher.init(Cipher.DECRYPT_MODE, key);
				byte[] senderBytes = Base64.getDecoder().decode(getSender());
				String sender = new String(cipher.doFinal(senderBytes));
				cipher.init(Cipher.DECRYPT_MODE, key);
				byte[] subjectBytes = Base64.getDecoder().decode(getSubject());
				String subject = new String(cipher.doFinal(subjectBytes));

				cipher.init(Cipher.DECRYPT_MODE, key);
				byte[] messageBytes = Base64.getDecoder().decode(getMessage());
				String message = new String(cipher.doFinal(messageBytes));

				Map<String, String> options = new HashMap<String, String>();
				for (String text : getOptionsTextActionMap().keySet()) {
					cipher.init(Cipher.DECRYPT_MODE, key);
					byte[] textBytes = Base64.getDecoder().decode(text);
					String k = new String(cipher.doFinal(textBytes));
					cipher.init(Cipher.DECRYPT_MODE, key);
					byte[] actionBytes = Base64.getDecoder().decode(getOptionsTextActionMap().get(text));
					String v = new String(cipher.doFinal(actionBytes));
					options.put(k, v);
				}

				return new MailEntry(getTimestamp(), sender, receiver, subject, message, options);
			} catch (IllegalBlockSizeException | BadPaddingException | InvalidKeyException e) {
				e.printStackTrace();
				return null;
			}
		}

		public String toString() {
			StringBuilder sb = new StringBuilder("MailEntry { ");
			
			sb.append("[timestamp: " + getTimestamp() + "], ");
			sb.append("[sender: " + getSender() + "], ");
			sb.append("[receiver: " + getReceiver() + "], ");
			sb.append("[subject: " + getSubject() + "], ");
			sb.append("[message: " + getMessage() + "], ");
			
			sb.append(" }");
			return sb.toString();
		}
	}

	private Path getDatabaseFilePath(String colleagueName) {
		return getDatabaseFilePath(colleagueName, false);
	}

	private Path getDatabaseFilePath(String colleagueName, boolean demoMode) {
		String deadDropsPath = Paths.get(FrameIO.PARENT_FOLDER).resolve("resources-" + user).resolve("deaddrops").toAbsolutePath().toString();//FrameIO.DEAD_DROPS_PATH;
		
//		if (deadDropsPath == null) {
//			deadDropsPath = Paths.get(FrameIO.PARENT_FOLDER).resolve("resources-" + colleagueName).resolve("deadrops").toAbsolutePath().toString();
//		} else if (demoMode) {
//			deadDropsPath = deadDropsPath.replace(user, colleagueName);
//		}
		
		if (demoMode) {
			deadDropsPath = deadDropsPath.replace(user, colleagueName);
		}

		Path databaseFileDirectoryPath = Paths.get(deadDropsPath).resolve(user + "+" + colleagueName);
		if (!databaseFileDirectoryPath.toFile().exists()) {
			databaseFileDirectoryPath = Paths.get(deadDropsPath).resolve(colleagueName + "+" + user);
		}
		
		if (!databaseFileDirectoryPath.toFile().exists()) {
			databaseFileDirectoryPath.toFile().mkdirs();
		}

		Path databaseFilePath = databaseFileDirectoryPath.resolve(colleagueName + ".db");
		if (!databaseFilePath.toFile().exists()) {
			createMailDatabaseFile(databaseFilePath);
		}
		return databaseFilePath;
	}

	private void createMailDatabaseFile(Path databaseFilePath) {
		databaseFilePath.getParent().toFile().mkdir();
		//@formatter:off
		String sql = 
				"CREATE TABLE EXPMAIL (" +
				"TIME		TEXT					NOT NULL, " +
				"SND		TEXT					NOT NULL, " + 
				"REC		TEXT					NOT NULL, " + 
				"SUB		TEXT					NOT NULL, " +
				"MSG		TEXT					NOT NULL, " + 
				"OPTS		ARRAY					NOT NULL, " +
				"OPTSVAL	ARRAY					NOT NULL)";
		//@formatter:on
		try (Connection c = DriverManager.getConnection("jdbc:sqlite:" + databaseFilePath.toAbsolutePath())) {
			Statement createTable = c.createStatement();
			createTable.executeUpdate(sql);
			createTable.close();
		} catch (SQLException e) {
			System.err.println("Error while creating database file.");
			e.printStackTrace();
		}
	}

	private PrivateKey getPrivateKey() {
		Text keyItem = KeyList.PrivateKey.get();
		String noKeyMessage = "No private key present: your communication with other Expeditee users will be limited until this is resolved.";
		if (keyItem.getData() != null) {
			// Check mail.
			String keyEncoded = keyItem.getData().get(0);
			byte[] keyBytes = Base64.getDecoder().decode(keyEncoded);
			try {
				return KeyFactory.getInstance(AsymmetricAlgorithm).generatePrivate(new PKCS8EncodedKeySpec(keyBytes));
			} catch (InvalidKeySpecException | NoSuchAlgorithmException e) {
				MessageBay.errorMessage(noKeyMessage);
				MessageBay.errorMessage("See exception stack trace for more information.");
				e.printStackTrace();
				return null;
			}
		} else {
			MessageBay.errorMessage(noKeyMessage);
			return null;
		}
	}

	private List<Path> getDatabaseInboxFiles() {
		List<Path> databaseFiles = new ArrayList<Path>();
		Path deadDropPath = Paths.get(FrameIO.DEAD_DROPS_PATH);
		File deadDropDirectory = deadDropPath.toFile();
		if (!deadDropDirectory.exists()) { return databaseFiles; }
		for (File partnershipDirectoryCanditate : deadDropDirectory.listFiles()) {
			if (partnershipDirectoryCanditate.isDirectory()) {
				Path partnershipDirectory = Paths.get(partnershipDirectoryCanditate.getAbsolutePath());
				Path dbFile = partnershipDirectory.resolve(user + ".db");
				if (dbFile.toFile().exists()) {
					databaseFiles.add(dbFile);
				}
			}
		}
		return databaseFiles;
	}

	private List<MailEntry> loadMailFromDatabases(List<Path> databasePaths) {
		List<MailEntry> mail = new ArrayList<MailEntry>();

		for (Path dbPath : databasePaths) {
			try {
				mail.addAll(loadMailFromDatabase(dbPath));
			} catch (SQLException e) {
				e.printStackTrace();
			}
		}

		return mail;
	}

	private List<MailEntry> loadMailFromDatabase(Path dbPath) throws SQLException {
		List<MailEntry> mail = new ArrayList<MailEntry>();

		Connection c = DriverManager.getConnection("jdbc:sqlite:" + dbPath.toAbsolutePath());
		String sql = "SELECT * FROM EXPMAIL";
		PreparedStatement query = c.prepareStatement(sql);
		ResultSet results = query.executeQuery();

		while (results.next()) {
			String timestamp = results.getString("TIME");
			
			Path lastAccessedPath = dbPath.getParent().resolve(user + ".last-accessed");
			try (Scanner in = new Scanner(lastAccessedPath.toFile())) {
				Date lastAccessedTimestamp = FORMAT.parse(in.nextLine());
				Date mailTimestamp = FORMAT.parse(timestamp);
				if (mailTimestamp.before(lastAccessedTimestamp)) {
					continue;
				}
			} catch (FileNotFoundException | ParseException e) {
				// It may not have been created yet, then err on the safe side and add it in.
				// If it fails to parse then we again err on the safe side.
			} catch (RuntimeException e) {
				e.printStackTrace();
				continue;
			}
			
			String from = results.getString("SND");
			String to = results.getString("REC");
			String subject = results.getString("SUB");
			String message = results.getString("MSG");
			String[] opts = results.getString("opts").split(",");
			opts[0] = opts[0].replace("[", "");
			opts[opts.length - 1] = opts[opts.length - 1].replace("]", "");
			String[] optsVal = results.getString("optsval").split(",");
			optsVal[0] = optsVal[0].replace("[", "");
			optsVal[optsVal.length - 1] = optsVal[optsVal.length - 1].replace("]", "");
			Map<String, String> options = new HashMap<String, String>();
			for (int i = 0; i < opts.length; i++) {
				String key = opts[i].trim();
				String val = optsVal[i].trim();
				options.put(key, val);
			}

			MailEntry mailEntry = new MailEntry(timestamp, from, to, subject, message, options);
			mailEntry.setDatabaseSource(dbPath);
			mail.add(mailEntry);
		}

		results.close();
		query.close();
		c.close();

		return mail;
	}

	private List<MailEntry> decryptMailEntries(List<MailEntry> encryptedMailMessages, PrivateKey key) {
		List<MailEntry> mail = new ArrayList<MailEntry>();

		for (MailEntry encMail : encryptedMailMessages) {
			if (encMail.isSingleUseEncryption()) {
				MailEntry oneOffSecureSubstitute = generateOneOffSecureSubstitute(encMail);
				if (oneOffSecureSubstitute != null) {
					mail.add(oneOffSecureSubstitute);
				}
			} else {
				MailEntry decryptMailEntry = encMail.decryptMailEntry(key, () -> {
					try {
						return Cipher.getInstance(AsymmetricAlgorithm + AsymmetricAlgorithmParameters);
					} catch (NoSuchAlgorithmException | NoSuchPaddingException e) {
						e.printStackTrace();
						return null;
					}
				}, "");
				if (decryptMailEntry != null) {
					mail.add(decryptMailEntry);
				}
			}
		}

		return mail;
	}

	private void touchLastAccessedFiles(List<Path> databasePaths) {
		for (Path databasePath : databasePaths) {
			Path lastAccessedFilePath = databasePath.getParent().resolve(user + ".last-accessed");
			try (FileWriter out = new FileWriter(lastAccessedFilePath.toFile())) {
				out.write(Mail.FORMAT.format(new Date()) + System.getProperty("line.separator"));
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}

	private MailEntry generateOneOffSecureSubstitute(MailEntry toWrap) {
		StringBuilder sb = new StringBuilder();
		String sep = ":::";
		sb.append("Read one-off secure message." + sep);
		sb.append(toWrap.timestamp + sep);
		sb.append(toWrap.sender.substring(1) + sep);
		sb.append(toWrap.receiver + sep);
		sb.append(toWrap.subject + sep);
		sb.append(toWrap.message + sep);
		for (String k: toWrap.options.keySet()) {
			sb.append(k + sep + toWrap.options.get(k) + sep);
		}
		sb.reverse().delete(0, 3).reverse();
		Map<String, String> options = new HashMap<String, String>();
		options.put(sb.toString(), "AuthOneOffSecureMessage");
		MailEntry mailOuter = new MailEntry(toWrap.timestamp, "Unknown", user, "You have received a one-off secure message.",
				"Check your email for more details.", options);
		return mailOuter;
	}

	public String getUser() {
		return user;
	}
}
